@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import glob
 import json
 import os
@@ -7,7 +9,6 @@ import subprocess
 import sys
 import sysconfig
 from distutils import log
-from distutils.command.build import build as CommandBuild
 from distutils.errors import (
     CompileError,
     DistutilsExecError,
@@ -15,22 +16,19 @@ from distutils.errors import (
     DistutilsPlatformError,
 )
 from distutils.sysconfig import get_config_var
-from subprocess import check_output
-from typing import Dict, List, NamedTuple, Optional, Union, cast
+from pathlib import Path
+from typing import Dict, Iterable, List, NamedTuple, Optional, Set, Tuple, cast
 
+import pkg_resources
+from setuptools.command.build import build as CommandBuild  # type: ignore[import]
 from setuptools.command.build_ext import build_ext as CommandBuildExt
 from setuptools.command.build_ext import get_abi3_suffix
 from typing_extensions import Literal
 
+from ._utils import format_called_process_error
 from .command import RustCommand
-from .extension import Binding, RustExtension, Strip
-from .utils import (
-    PyLimitedApi,
-    binding_features,
-    get_rust_target_info,
-    get_rust_target_list,
-    split_platform_and_extension,
-)
+from .extension import Binding, RustBin, RustExtension, Strip
+from .rustc_info import get_rust_host, get_rust_target_list, get_rustc_cfgs
 
 
 class build_rust(RustCommand):
@@ -57,7 +55,7 @@ class build_rust(RustCommand):
     ]
     boolean_options = ["inplace", "debug", "release", "qbuild"]
 
-    plat_name: Optional[str]
+    plat_name: Optional[str] = None
 
     def initialize_options(self) -> None:
         super().initialize_options()
@@ -67,14 +65,17 @@ class build_rust(RustCommand):
         self.qbuild = None
         self.build_temp = None
         self.plat_name = None
+        self.build_number = None
         self.target = os.getenv("CARGO_BUILD_TARGET")
         self.cargo = os.getenv("CARGO", "cargo")
 
     def finalize_options(self) -> None:
         super().finalize_options()
 
+        self.data_dir = self.get_data_dir()
+
         if self.plat_name is None:
-            self.plat_name = cast(
+            self.plat_name = cast(  # type: ignore[no-any-unimported]
                 CommandBuild, self.get_finalized_command("build")
             ).plat_name
             assert isinstance(self.plat_name, str)
@@ -87,6 +88,18 @@ class build_rust(RustCommand):
             ("inplace", "inplace"),
         )
 
+        if self.build_number is not None and not self.build_number[:1].isdigit():
+            raise ValueError("Build tag (build-number) must start with a digit.")
+
+    def get_data_dir(self) -> str:
+        components = (
+            pkg_resources.safe_name(self.distribution.get_name()).replace("-", "_"),  # type: ignore[attr-defined]
+            pkg_resources.safe_version(self.distribution.get_version()).replace("-", "_"),  # type: ignore[attr-defined]
+        )
+        if self.build_number:
+            components += (self.build_number,)
+        return "-".join(components) + ".data"
+
     def run_for_extension(self, ext: RustExtension) -> None:
         assert self.plat_name is not None
 
@@ -94,6 +107,12 @@ class build_rust(RustCommand):
         universal2 = False
         if self.plat_name.startswith("macosx-") and arch_flags:
             universal2 = "x86_64" in arch_flags and "arm64" in arch_flags
+            if not universal2 and not self.target:
+                if "arm64" in arch_flags:
+                    self.target = "aarch64-apple-darwin"
+                elif "x86_64" in arch_flags:
+                    self.target = "x86_64-apple-darwin"
+
         if universal2:
             arm64_dylib_paths = self.build_extension(ext, "aarch64-apple-darwin")
             x86_64_dylib_paths = self.build_extension(ext, "x86_64-apple-darwin")
@@ -109,60 +128,47 @@ class build_rust(RustCommand):
         self.install_extension(ext, dylib_paths)
 
     def build_extension(
-        self, ext: RustExtension, forced_target_triple: Optional[str]
+        self, ext: RustExtension, forced_target_triple: Optional[str] = None
     ) -> List["_BuiltModule"]:
 
-        target_info = self._detect_rust_target(forced_target_triple)
-        if target_info is not None:
-            target_triple = target_info.triple
-            cross_lib = target_info.cross_lib
-            linker = target_info.linker
-            # We're ignoring target_info.linker_args for now because we're not
-            # sure if they will always do the right thing. Might help with some
-            # of the OS-specific logic if it does.
+        target_triple = self._detect_rust_target(forced_target_triple)
+        rustc_cfgs = get_rustc_cfgs(target_triple)
 
-        else:
-            target_triple = None
-            cross_lib = None
-            linker = None
-
-        rustc_cfgs = _get_rustc_cfgs(target_triple)
-
-        env = _prepare_build_environment(cross_lib)
+        env = _prepare_build_environment()
 
         if not os.path.exists(ext.path):
             raise DistutilsFileError(
                 f"can't find Rust extension project file: {ext.path}"
             )
 
-        # Find where to put the temporary build files created by `cargo`
-        target_dir = _base_cargo_target_dir(ext)
-        if target_triple is not None:
-            target_dir = os.path.join(target_dir, target_triple)
-
         quiet = self.qbuild or ext.quiet
         debug = self._is_debug_build(ext)
+
         cargo_args = self._cargo_args(
             ext=ext, target_triple=target_triple, release=not debug, quiet=quiet
         )
 
+        rustflags = []
+
         if ext._uses_exec_binding():
-            command = [self.cargo, "build", "--manifest-path", ext.path, *cargo_args]
+            command = [
+                self.cargo,
+                "build",
+                "--manifest-path",
+                ext.path,
+                "--message-format=json-render-diagnostics",
+                *cargo_args,
+            ]
 
         else:
             rustc_args = [
                 "--crate-type",
                 "cdylib",
+                *ext.rustc_flags,
             ]
 
-            if ext.rustc_flags is not None:
-                rustc_args.extend(ext.rustc_flags)
-
-            if linker is not None:
-                rustc_args.extend(["-C", "linker=" + linker])
-
             # OSX requires special linker arguments
-            if sys.platform == "darwin":
+            if rustc_cfgs.get("target_os") == "macos":
                 ext_basename = os.path.basename(self.get_dylib_ext_path(ext, ext.name))
                 rustc_args.extend(
                     [
@@ -171,29 +177,24 @@ class build_rust(RustCommand):
                     ]
                 )
 
-            if ext.native:
-                rustc_args.extend(["-C", "target-cpu=native"])
-
             # Tell musl targets not to statically link libc. See
             # https://github.com/rust-lang/rust/issues/59302 for details.
             if rustc_cfgs.get("target_env") == "musl":
                 # This must go in the env otherwise rustc will refuse to build
                 # the cdylib, see https://github.com/rust-lang/cargo/issues/10143
-                MUSL_FLAGS = "-C target-feature=-crt-static"
-                rustflags = env.get("RUSTFLAGS")
-                if rustflags is not None:
-                    env["RUSTFLAGS"] = f"{rustflags} {MUSL_FLAGS}"
-                else:
-                    env["RUSTFLAGS"] = MUSL_FLAGS
+                rustflags.append("-Ctarget-feature=-crt-static")
 
-                # Include this in the command-line anyway, so that when verbose
-                # logging enabled the user will see that this flag is in use.
-                rustc_args.extend(MUSL_FLAGS.split())
+            elif (rustc_cfgs.get("target_arch"), rustc_cfgs.get("target_os")) == (
+                "wasm32",
+                "emscripten",
+            ):
+                rustc_args.extend(["-C", f"link-args=-sSIDE_MODULE=2 -sWASM_BIGINT"])
 
             command = [
                 self.cargo,
                 "rustc",
                 "--lib",
+                "--message-format=json-render-diagnostics",
                 "--manifest-path",
                 ext.path,
                 *cargo_args,
@@ -201,14 +202,35 @@ class build_rust(RustCommand):
                 *rustc_args,
             ]
 
+        if rustflags:
+            existing_rustflags = env.get("RUSTFLAGS")
+            if existing_rustflags is not None:
+                rustflags.append(existing_rustflags)
+            new_rustflags = " ".join(rustflags)
+            env["RUSTFLAGS"] = new_rustflags
+
+            # print RUSTFLAGS being added before the command
+            if not quiet:
+                print(f"[RUSTFLAGS={new_rustflags}]", end=" ", file=sys.stderr)
+
         if not quiet:
             print(" ".join(command), file=sys.stderr)
 
         # Execute cargo
         try:
-            output = subprocess.check_output(command, env=env, encoding="latin-1")
+            # If quiet, capture all output and only show it in the exception
+            # If not quiet, forward all cargo output to stderr
+            stderr = subprocess.PIPE if quiet else None
+            cargo_messages = subprocess.check_output(
+                command,
+                env=env,
+                stderr=stderr,
+                text=True,
+            )
         except subprocess.CalledProcessError as e:
-            raise CompileError(f"cargo failed with code: {e.returncode}\n{e.output}")
+            # Don't include stdout in the formatted error as it is a huge dump
+            # of cargo json lines which aren't helpful for the end user.
+            raise CompileError(format_called_process_error(e, include_stdout=False))
 
         except OSError:
             raise DistutilsExecError(
@@ -216,53 +238,65 @@ class build_rust(RustCommand):
                 "requires Rust to be installed and cargo to be on the PATH"
             )
 
-        if not quiet:
-            if output:
-                print(output, file=sys.stderr)
-
         # Find the shared library that cargo hopefully produced and copy
         # it into the build directory as if it were produced by build_ext.
 
-        artifacts_dir = os.path.join(target_dir, "debug" if debug else "release")
         dylib_paths = []
+        package_id = ext.metadata(quiet=quiet)["resolve"]["root"]
 
         if ext._uses_exec_binding():
+            # Find artifact from cargo messages
+            artifacts = _find_cargo_artifacts(
+                cargo_messages.splitlines(),
+                package_id=package_id,
+                kind="bin",
+            )
             for name, dest in ext.target.items():
                 if not name:
                     name = dest.split(".")[-1]
-                exe = sysconfig.get_config_var("EXE")
-                if exe is not None:
-                    name += exe
 
-                path = os.path.join(artifacts_dir, name)
-                if os.access(path, os.X_OK):
-                    dylib_paths.append(_BuiltModule(dest, path))
-                else:
+                try:
+                    artifact_path = next(
+                        artifact
+                        for artifact in artifacts
+                        if Path(artifact).with_suffix("").name == name
+                    )
+                except StopIteration:
                     raise DistutilsExecError(
-                        "Rust build failed; "
-                        f"unable to find executable '{name}' in '{artifacts_dir}'"
+                        f"Rust build failed; unable to locate executable '{name}'"
                     )
+
+                if os.environ.get("CARGO") == "cross":
+                    artifact_path = _replace_cross_target_dir(
+                        artifact_path, ext, quiet=quiet
+                    )
+
+                dylib_paths.append(_BuiltModule(dest, artifact_path))
         else:
-            if sys.platform == "win32" or sys.platform == "cygwin":
-                dylib_ext = "dll"
-            elif sys.platform == "darwin":
-                dylib_ext = "dylib"
-            else:
-                dylib_ext = "so"
-
-            wildcard_so = "*{}.{}".format(ext.get_lib_name(), dylib_ext)
-
-            try:
-                dylib_paths.append(
-                    _BuiltModule(
-                        ext.name,
-                        next(glob.iglob(os.path.join(artifacts_dir, wildcard_so))),
-                    )
-                )
-            except StopIteration:
+            # Find artifact from cargo messages
+            artifacts = _find_cargo_artifacts(
+                cargo_messages.splitlines(),
+                package_id=package_id,
+                kind="cdylib",
+            )
+            if len(artifacts) == 0:
                 raise DistutilsExecError(
-                    f"Rust build failed; unable to find any {wildcard_so} in {artifacts_dir}"
+                    "Rust build failed; unable to find any build artifacts"
                 )
+            elif len(artifacts) > 1:
+                raise DistutilsExecError(
+                    f"Rust build failed; expected only one build artifact but found {artifacts}"
+                )
+
+            artifact_path = artifacts[0]
+
+            if os.environ.get("CARGO") == "cross":
+                artifact_path = _replace_cross_target_dir(
+                    artifact_path, ext, quiet=quiet
+                )
+
+            # guaranteed to be just one element after checks above
+            dylib_paths.append(_BuiltModule(ext.name, artifact_path))
         return dylib_paths
 
     def install_extension(
@@ -287,7 +321,7 @@ class build_rust(RustCommand):
             if ext._uses_exec_binding():
                 ext_path = build_ext.get_ext_fullpath(module_name)
                 # remove extensions
-                ext_path, _, _ = split_platform_and_extension(ext_path)
+                ext_path, _, _ = _split_platform_and_extension(ext_path)
 
                 # Add expected extension
                 exe = sysconfig.get_config_var("EXE")
@@ -295,7 +329,17 @@ class build_rust(RustCommand):
                     ext_path += exe
 
                 os.makedirs(os.path.dirname(ext_path), exist_ok=True)
-                ext.install_script(module_name.split(".")[-1], ext_path)
+                if isinstance(ext, RustBin):
+                    executable_name = module_name
+                    if exe is not None:
+                        executable_name += exe
+                    scripts_dir = os.path.join(
+                        build_ext.build_lib, self.data_dir, "scripts"
+                    )
+                    os.makedirs(scripts_dir, exist_ok=True)
+                    ext_path = os.path.join(scripts_dir, executable_name)
+                else:
+                    ext.install_script(module_name.split(".")[-1], ext_path)
             else:
                 ext_path = self.get_dylib_ext_path(ext, module_name)
                 os.makedirs(os.path.dirname(ext_path), exist_ok=True)
@@ -359,12 +403,12 @@ class build_rust(RustCommand):
         host_arch = host_platform.rsplit("-", 1)[1]
         # Remove incorrect platform tag if we are cross compiling
         if target_arch and host_arch != target_arch:
-            ext_path, _, extension = split_platform_and_extension(ext_path)
+            ext_path, _, extension = _split_platform_and_extension(ext_path)
             # rust.so, removed platform tag
             ext_path += extension
         return ext_path
 
-    def _py_limited_api(self) -> PyLimitedApi:
+    def _py_limited_api(self) -> _PyLimitedApi:
         bdist_wheel = self.distribution.get_command_obj("bdist_wheel", create=False)
 
         if bdist_wheel is None:
@@ -375,72 +419,30 @@ class build_rust(RustCommand):
 
             bdist_wheel_command = cast(CommandBdistWheel, bdist_wheel)  # type: ignore[no-any-unimported]
             bdist_wheel_command.ensure_finalized()
-            return cast(PyLimitedApi, bdist_wheel_command.py_limited_api)
+            return cast(_PyLimitedApi, bdist_wheel_command.py_limited_api)
 
     def _detect_rust_target(
-        self, forced_target_triple: Optional[str]
-    ) -> Optional["_TargetInfo"]:
-        cross_compile_info = _detect_unix_cross_compile_info()
-        if cross_compile_info is not None:
-            cross_target_info = cross_compile_info.to_target_info()
-            if forced_target_triple is not None:
-                if (
-                    cross_target_info is not None
-                    and not cross_target_info.is_compatible_with(forced_target_triple)
-                ):
-                    self.warn(
-                        f"Forced Rust target `{forced_target_triple}` is not "
-                        f"compatible with deduced Rust target "
-                        f"`{cross_target_info.triple}` - the built package "
-                        f" may not import successfully once installed."
-                    )
-
-                # Forcing the target in a cross-compile environment; use
-                # the cross-compile information in combination with the
-                # forced target
-                return _TargetInfo(
-                    forced_target_triple,
-                    cross_compile_info.cross_lib,
-                    cross_compile_info.linker,
-                    cross_compile_info.linker_args,
-                )
-            elif cross_target_info is not None:
-                return cross_target_info
-            else:
-                raise DistutilsPlatformError(
-                    "Don't know the correct rust target for system type "
-                    f"{cross_compile_info.host_type}. Please set the "
-                    "CARGO_BUILD_TARGET environment variable."
-                )
-
-        elif forced_target_triple is not None:
-            return _TargetInfo.for_triple(forced_target_triple)
-
-        else:
+        self, forced_target_triple: Optional[str] = None
+    ) -> Optional[str]:
+        assert self.plat_name is not None
+        if forced_target_triple is not None:
             # Automatic target detection can be overridden via the CARGO_BUILD_TARGET
             # environment variable or --target command line option
-            return self._detect_local_rust_target()
+            return forced_target_triple
 
-    def _detect_local_rust_target(self) -> Optional["_TargetInfo"]:
-        """Attempts to infer the correct Rust target from build environment for
-        some edge cases."""
-        assert self.plat_name is not None
+        # Determine local rust target which needs to be "forced" if necessary
+        local_rust_target = _adjusted_local_rust_target(self.plat_name)
 
-        # If we are on a 64-bit machine, but running a 32-bit Python, then
-        # we'll target a 32-bit Rust build.
-        if self.plat_name == "win32":
-            if _get_rustc_cfgs(None).get("target_env") == "gnu":
-                return _TargetInfo.for_triple("i686-pc-windows-gnu")
-            return _TargetInfo.for_triple("i686-pc-windows-msvc")
-        elif self.plat_name == "win-amd64":
-            if _get_rustc_cfgs(None).get("target_env") == "gnu":
-                return _TargetInfo.for_triple("x86_64-pc-windows-gnu")
-            return _TargetInfo.for_triple("x86_64-pc-windows-msvc")
-        elif self.plat_name.startswith("macosx-") and platform.machine() == "x86_64":
-            # x86_64 or arm64 macOS targeting x86_64
-            return _TargetInfo.for_triple("x86_64-apple-darwin")
-        else:
-            return None
+        # Match cargo's behaviour of not using an explicit target if the
+        # target we're compiling for is the host
+        if (
+            local_rust_target is not None
+            # check for None first to avoid calling to rustc if not needed
+            and local_rust_target != get_rust_host()
+        ):
+            return local_rust_target
+
+        return None
 
     def _is_debug_build(self, ext: RustExtension) -> bool:
         if self.release:
@@ -464,7 +466,9 @@ class build_rust(RustCommand):
             args.extend(["--target", target_triple])
 
         if release:
-            args.append("--release")
+            profile = ext.get_cargo_profile()
+            if not profile:
+                args.append("--release")
 
         if quiet:
             args.append("-q")
@@ -476,7 +480,7 @@ class build_rust(RustCommand):
 
         features = {
             *ext.features,
-            *binding_features(ext, py_limited_api=self._py_limited_api()),
+            *_binding_features(ext, py_limited_api=self._py_limited_api()),
         }
 
         if features:
@@ -485,6 +489,9 @@ class build_rust(RustCommand):
         if ext.args is not None:
             args.extend(ext.args)
 
+        if ext.cargo_manifest_args is not None:
+            args.extend(ext.cargo_manifest_args)
+
         return args
 
 
@@ -492,11 +499,9 @@ def create_universal2_binary(output_path: str, input_paths: List[str]) -> None:
     # Try lipo first
     command = ["lipo", "-create", "-output", output_path, *input_paths]
     try:
-        subprocess.check_output(command)
+        subprocess.check_output(command, text=True)
     except subprocess.CalledProcessError as e:
         output = e.output
-        if isinstance(output, bytes):
-            output = e.output.decode("latin-1").strip()
         raise CompileError("lipo failed with code: %d\n%s" % (e.returncode, output))
     except OSError:
         # lipo not found, try using the fat-macho library
@@ -525,106 +530,6 @@ class _BuiltModule(NamedTuple):
     path: str
 
 
-class _TargetInfo(NamedTuple):
-    triple: str
-    cross_lib: Optional[str]
-    linker: Optional[str]
-    linker_args: Optional[str]
-
-    @staticmethod
-    def for_triple(triple: str) -> "_TargetInfo":
-        return _TargetInfo(triple, None, None, None)
-
-    def is_compatible_with(self, target: str) -> bool:
-        if self.triple == target:
-            return True
-
-        # the vendor field can be ignored, so x86_64-pc-linux-gnu is compatible
-        # with x86_64-unknown-linux-gnu
-        if _replace_vendor_with_unknown(self.triple) == target:
-            return True
-
-        return False
-
-
-class _CrossCompileInfo(NamedTuple):
-    host_type: str
-    cross_lib: Optional[str]
-    linker: Optional[str]
-    linker_args: Optional[str]
-
-    def to_target_info(self) -> Optional[_TargetInfo]:
-        """Maps this cross compile info to target info.
-
-        Returns None if the corresponding target information could not be
-        deduced.
-        """
-        # hopefully an exact match
-        targets = get_rust_target_list()
-        if self.host_type in targets:
-            return _TargetInfo(
-                self.host_type, self.cross_lib, self.linker, self.linker_args
-            )
-
-        # the vendor field can be ignored, so x86_64-pc-linux-gnu is compatible
-        # with x86_64-unknown-linux-gnu
-        without_vendor = _replace_vendor_with_unknown(self.host_type)
-        if without_vendor is not None and without_vendor in targets:
-            return _TargetInfo(
-                without_vendor, self.cross_lib, self.linker, self.linker_args
-            )
-
-        return None
-
-
-def _detect_unix_cross_compile_info() -> Optional["_CrossCompileInfo"]:
-    # See https://github.com/PyO3/setuptools-rust/issues/138
-    # This is to support cross compiling on *NIX, where plat_name isn't
-    # necessarily the same as the system we are running on.  *NIX systems
-    # have more detailed information available in sysconfig. We need that
-    # because plat_name doesn't give us information on e.g., glibc vs musl.
-    host_type = sysconfig.get_config_var("HOST_GNU_TYPE")
-    build_type = sysconfig.get_config_var("BUILD_GNU_TYPE")
-
-    if not host_type or host_type == build_type:
-        # not *NIX, or not cross compiling
-        return None
-
-    if "apple-darwin" in host_type and (build_type and "apple-darwin" in build_type):
-        # On macos and the build and host differ. This is probably an arm
-        # Python which was built on x86_64. Don't try to handle this for now.
-        # (See https://github.com/PyO3/setuptools-rust/issues/192)
-        return None
-
-    stdlib = sysconfig.get_path("stdlib")
-    assert stdlib is not None
-    cross_lib = os.path.dirname(stdlib)
-
-    bldshared = sysconfig.get_config_var("BLDSHARED")
-    if not bldshared:
-        linker = None
-        linker_args = None
-    else:
-        [linker, linker_args] = bldshared.split(maxsplit=1)
-
-    return _CrossCompileInfo(host_type, cross_lib, linker, linker_args)
-
-
-_RustcCfgs = Dict[str, Optional[str]]
-
-
-def _get_rustc_cfgs(target_triple: Optional[str]) -> _RustcCfgs:
-    cfgs: _RustcCfgs = {}
-    for entry in get_rust_target_info(target_triple):
-        maybe_split = entry.split("=", maxsplit=1)
-        if len(maybe_split) == 2:
-            cfgs[maybe_split[0]] = maybe_split[1].strip('"')
-        else:
-            assert len(maybe_split) == 1
-            cfgs[maybe_split[0]] = None
-    return cfgs
-
-
 def _replace_vendor_with_unknown(target: str) -> Optional[str]:
     """Replaces vendor in the target triple with unknown.
 
@@ -637,7 +542,7 @@ def _replace_vendor_with_unknown(target: str) -> Optional[str]:
     return "-".join(components)
 
 
-def _prepare_build_environment(cross_lib: Optional[str]) -> Dict[str, str]:
+def _prepare_build_environment() -> Dict[str, str]:
     """Prepares environment variables to use when executing cargo build."""
 
     # Make sure that if pythonXX-sys is used, it builds against the current
@@ -657,29 +562,12 @@ def _prepare_build_environment(cross_lib: Optional[str]) -> Dict[str, str]:
             "PYO3_PYTHON": os.environ.get("PYO3_PYTHON", sys.executable),
         }
     )
-
-    if cross_lib:
-        env.setdefault("PYO3_CROSS_LIB_DIR", cross_lib)
-
     return env
 
 
-def _base_cargo_target_dir(ext: RustExtension) -> str:
-    """Returns the root target directory cargo will use.
-
-    If --target is passed to cargo in the command line, the target directory
-    will have the target appended as a child.
-    """
-    target_directory = ext._metadata()["target_directory"]
-    assert isinstance(
-        target_directory, str
-    ), "expected cargo metadata to contain a string target directory"
-    return target_directory
-
-
 def _is_py_limited_api(
     ext_setting: Literal["auto", True, False],
-    wheel_setting: Optional[PyLimitedApi],
+    wheel_setting: Optional[_PyLimitedApi],
 ) -> bool:
     """Returns whether this extension is being built for the limited api.
 
@@ -702,3 +590,127 @@ def _is_py_limited_api(
 
     # "auto" setting - use whether the bdist_wheel option is truthy.
     return bool(wheel_setting)
+
+
+def _binding_features(
+    ext: RustExtension,
+    py_limited_api: _PyLimitedApi,
+) -> Set[str]:
+    if ext.binding in (Binding.NoBinding, Binding.Exec):
+        return set()
+    elif ext.binding is Binding.PyO3:
+        features = {"pyo3/extension-module"}
+        if ext.py_limited_api == "auto":
+            if isinstance(py_limited_api, str):
+                python_version = py_limited_api[2:]
+                features.add(f"pyo3/abi3-py{python_version}")
+            elif py_limited_api:
+                features.add(f"pyo3/abi3")
+        return features
+    elif ext.binding is Binding.RustCPython:
+        return {"cpython/python3-sys", "cpython/extension-module"}
+    else:
+        raise DistutilsPlatformError(f"unknown Rust binding: '{ext.binding}'")
+
+
+_PyLimitedApi = Literal["cp37", "cp38", "cp39", "cp310", "cp311", "cp312", True, False]
+
+
+def _adjusted_local_rust_target(plat_name: str) -> Optional[str]:
+    """Returns the local rust target for the given `plat_name`, if it is
+    necessary to 'force' a specific target for correctness."""
+
+    # If we are on a 64-bit machine, but running a 32-bit Python, then
+    # we'll target a 32-bit Rust build.
+    if plat_name == "win32":
+        if get_rustc_cfgs(None).get("target_env") == "gnu":
+            return "i686-pc-windows-gnu"
+        else:
+            return "i686-pc-windows-msvc"
+    elif plat_name == "win-amd64":
+        if get_rustc_cfgs(None).get("target_env") == "gnu":
+            return "x86_64-pc-windows-gnu"
+        else:
+            return "x86_64-pc-windows-msvc"
+    elif plat_name.startswith("macosx-") and platform.machine() == "x86_64":
+        # x86_64 or arm64 macOS targeting x86_64
+        return "x86_64-apple-darwin"
+
+    return None
+
+
+def _split_platform_and_extension(ext_path: str) -> Tuple[str, str, str]:
+    """Splits an extension path into a tuple (ext_path, plat_tag, extension).
+
+    >>> _split_platform_and_extension("foo/bar.platform.so")
+    ('foo/bar', '.platform', '.so')
+    """
+
+    # rust.cpython-38-x86_64-linux-gnu.so to (rust.cpython-38-x86_64-linux-gnu, .so)
+    ext_path, extension = os.path.splitext(ext_path)
+    # rust.cpython-38-x86_64-linux-gnu to (rust, .cpython-38-x86_64-linux-gnu)
+    ext_path, platform_tag = os.path.splitext(ext_path)
+    return (ext_path, platform_tag, extension)
+
+
+def _find_cargo_artifacts(
+    cargo_messages: List[str],
+    *,
+    package_id: str,
+    kind: str,
+) -> List[str]:
+    """Identifies cargo artifacts built for the given `package_id` from the
+    provided cargo_messages.
+
+    >>> _find_cargo_artifacts(
+    ...    [
+    ...        '{"some_irrelevant_message": []}',
+    ...        '{"reason":"compiler-artifact","package_id":"some_id","target":{"kind":["cdylib"]},"filenames":["/some/path/baz.so"]}',
+    ...        '{"reason":"compiler-artifact","package_id":"some_id","target":{"kind":["cdylib", "rlib"]},"filenames":["/file/two/baz.dylib", "/file/two/baz.rlib"]}',
+    ...        '{"reason":"compiler-artifact","package_id":"some_other_id","target":{"kind":["cdylib"]},"filenames":["/not/this.so"]}',
+    ...    ],
+    ...    package_id="some_id",
+    ...    kind="cdylib",
+    ... )
+    ['/some/path/baz.so', '/file/two/baz.dylib']
+    >>> _find_cargo_artifacts(
+    ...    [
+    ...        '{"some_irrelevant_message": []}',
+    ...        '{"reason":"compiler-artifact","package_id":"some_id","target":{"kind":["cdylib"]},"filenames":["/some/path/baz.so"]}',
+    ...        '{"reason":"compiler-artifact","package_id":"some_id","target":{"kind":["cdylib", "rlib"]},"filenames":["/file/two/baz.dylib", "/file/two/baz.rlib"]}',
+    ...        '{"reason":"compiler-artifact","package_id":"some_other_id","target":{"kind":["cdylib"]},"filenames":["/not/this.so"]}',
+    ...    ],
+    ...    package_id="some_id",
+    ...    kind="rlib",
+    ... )
+    ['/file/two/baz.rlib']
+    """
+    artifacts = []
+    for message in cargo_messages:
+        # only bother parsing messages that look like a match
+        if "compiler-artifact" in message and package_id in message and kind in message:
+            parsed = json.loads(message)
+            # verify the message is correct
+            if (
+                parsed.get("reason") == "compiler-artifact"
+                and parsed.get("package_id") == package_id
+            ):
+                for artifact_kind, filename in zip(
+                    parsed["target"]["kind"], parsed["filenames"]
+                ):
+                    if artifact_kind == kind:
+                        artifacts.append(filename)
+    return artifacts
+
+
+def _replace_cross_target_dir(path: str, ext: RustExtension, *, quiet: bool) -> str:
+    """Replaces target director from `cross` docker build with the correct
+    local path.
+
+    Cross artifact messages and metadata contain paths from inside the
+    dockerfile; invoking `cargo metadata` we can work out the correct local
+    target directory.
+    """
+    cross_target_dir = ext._metadata(cargo="cross", quiet=quiet)["target_directory"]
+    local_target_dir = ext._metadata(cargo="cargo", quiet=quiet)["target_directory"]
+    return path.replace(cross_target_dir, local_target_dir)
